import os
import time
import logging
from typing import Dict
from AnyQt.QtCore import Qt

from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.canvas.mainwindow import OASYSMainWindow as _MainWindow, QDialog
    from orangecanvas.registry import set_global_registry
    from orangecanvas.registry.qt import QtWidgetRegistry
    from oasys.canvas import conf as orangeconfig
    from orangecanvas import config as canvasconfig

    class MainWindow(_MainWindow):
        def show_scheme_properties_for(self, scheme, window_title=None):
            return QDialog.Accepted

    try:
        from oasys.canvas.mainwindow import MainWindowRegistry
    except ImportError:
        MainWindowRegistry = None

else:
    from orangecanvas.registry import set_global_registry
    from orangecanvas.registry.qt import QtWidgetRegistry
    from orangecanvas import config as canvasconfig

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        # load MainWindow and config from Orange if installed
        from Orange.canvas.mainwindow import MainWindow
        from Orange.canvas import config as orangeconfig
    else:
        # else use the base one from orangewidget
        from orangewidget.workflow.mainwindow import OWCanvasMainWindow as MainWindow
        from ewoksorange.canvas import config as orangeconfig

from .utils import get_orange_canvas
from ..bindings import qtapp
from ..bindings.bindings import ows_file_context
from ..bindings.owwidgets import OWEwoksBaseWidget
from ..bindings.owsignal_manager import SignalManagerWithOutputTracking


_logger = logging.getLogger(__name__)


class OrangeCanvasHandler:
    """Run orange widget-based workflow manually (i.e. without executing the Qt application)"""

    def __init__(self):
        self.canvas = get_orange_canvas()
        self.__is_owner = self.canvas is None

    def __del__(self):
        self.close()

    def __enter__(self):
        if self.canvas is None:
            self._init_canvas()
            self.__is_owner = True
        return self

    def __exit__(self, *args):
        self.close()

    def _init_canvas(self):
        qtapp.ensure_qtapp()

        widget_registry = QtWidgetRegistry()
        set_global_registry(widget_registry)

        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            config = orangeconfig.oasysconf()
            config.init()
            canvasconfig.set_default(config)
            widget_discovery = config.widget_discovery(widget_registry)
            widget_discovery.run(config.widgets_entry_points())
        else:
            config = orangeconfig.Config()
            config.init()
            canvasconfig.set_default(config)
            widget_discovery = config.widget_discovery(widget_registry)
            widget_discovery.run(orangeconfig.widgets_entry_points())

        canvas = MainWindow()
        canvas.setAttribute(Qt.WA_DeleteOnClose)
        canvas.set_widget_registry(widget_registry)  # makes a copy of the registry

        if (
            ORANGE_VERSION == ORANGE_VERSION.oasys_fork
            and MainWindowRegistry is not None
        ):
            MainWindowRegistry.Instance().register_instance(
                instance=canvas, application_name=str(os.getpid())
            )  # need it for finding the canvas from the widgets

        self.canvas = canvas
        self.process_events()

    def close(self, force=False):
        if self.canvas is None or (not self.__is_owner and not force):
            return
        canvas, self.canvas = self.canvas, None
        self.process_events()
        # do not prompt for saving modification:
        canvas.current_document().setModified(False)
        canvas.close()
        self.process_events()

    def load_graph(self, graph, **kwargs):
        with ows_file_context(graph, **kwargs) as filename:
            self.load_ows(filename)

    def load_ows(self, filename: str):
        self.canvas.load_scheme(filename)

    @property
    def scheme(self):
        return self.canvas.current_document().scheme()

    @property
    def signal_manager(self) -> SignalManagerWithOutputTracking:
        signal_manager = self.scheme.signal_manager
        assert isinstance(
            signal_manager, SignalManagerWithOutputTracking
        ), "Orange signal manager was not patched before instantiated"
        return signal_manager

    def iter_nodes(self):
        for node in self.scheme.nodes:
            yield node

    def process_events(self):
        qtapp.process_qtapp_events()

    def show(self):
        qtapp.process_qtapp_events()
        self.canvas.show()
        qtapp.get_qtapp().exec()

    def widgets_from_name(self, name: str):
        for node in self.iter_nodes():
            if node.title == name:
                yield self.scheme.widget_for_node(node)

    def widget_from_id(self, id: str):
        return self.scheme.widget_for_node(self.scheme.nodes[int(id)])

    def iter_widgets(self):
        for node in self.iter_nodes():
            yield self.scheme.widget_for_node(node)

    def iter_widgets_with_name(self):
        for node in self.iter_nodes():
            yield node.title, self.scheme.widget_for_node(node)

    def iter_output_values(self):
        for name, widget in self.iter_widgets_with_name():
            yield name, widget.get_task_output_values()

    def get_output_values(self) -> Dict[str, dict]:
        return dict(self.iter_output_values())

    def set_input_values(self, inputs: Dict[str, dict]) -> None:
        for name, widget in self.iter_widgets_with_name():
            for adict in inputs:
                if adict["label"] == name:
                    widget.update_default_inputs(**{adict["name"]: adict["value"]})

    def start_workflow(self):
        triggered = False
        for node in self.iter_nodes():
            if not any(self.scheme.find_links(sink_node=node)):
                widget = self.scheme.widget_for_node(node)
                triggered = True
                _logger.info("Trigger workflow node %r", node.title)
                widget.handleNewSignals()
        if not triggered:
            _logger.warning("This workflow has no widgets that can be triggered")

    def wait_widgets(self, timeout=None, raise_error: bool = True):
        """Wait for all widgets to be executed. Widget failures are re-raised."""
        signal_manager = self.signal_manager
        widgets = list(self.iter_widgets())
        t0 = time.time()
        while True:
            self.process_events()
            executed = list()
            for widget in widgets:
                if raise_error and isinstance(widget, OWEwoksBaseWidget):
                    exception = widget.task_exception or widget.post_task_exception
                    if exception is not None:
                        raise exception
                is_executed = signal_manager.widget_is_executed(widget)
                executed.append(is_executed)
            if all(executed):
                break
            if timeout is not None:
                if (time.time() - t0) > timeout:
                    raise TimeoutError(timeout)
            time.sleep(0.1)



"""Copy parts of Orange.canvas.config to be used when Orange3 is not installed.
"""

from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.canvas.conf import oasysconf as _Config
    from oasys.canvas.conf import WIDGETS_ENTRY  # "oasys.widgets"
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.canvas.config import Config as _Config
    from Orange.canvas.config import WIDGETS_ENTRY  # "orange.widgets"
else:
    from orangewidget.workflow.config import Config as _Config
    from orangewidget.workflow.config import WIDGETS_ENTRY  # "orange.widgets"

from ..pkg_meta import iter_entry_points

EXAMPLE_WORKFLOWS_ENTRY = WIDGETS_ENTRY + ".tutorials"


class Config(_Config):
    @staticmethod
    def widgets_entry_points():
        """Return an `EntryPoint` iterator for all WIDGETS_ENTRY entry points."""
        # Ensure the 'this' distribution's ep is the first. iter_entry_points
        # yields them in unspecified order.
        from orangecontrib.ewokstest import is_ewokstest_category_enabled

        for ep in iter_entry_points(group=WIDGETS_ENTRY):
            if (
                _get_ep_module(ep) == "orangecontrib.ewokstest"
                and not is_ewokstest_category_enabled()
            ):
                continue
            yield ep

    @staticmethod
    def examples_entry_points():
        """Return an `EntryPoint` iterator for all EXAMPLE_WORKFLOWS_ENTRY entry points."""
        from orangecontrib.ewokstest import is_ewokstest_category_enabled

        for ep in iter_entry_points(group=EXAMPLE_WORKFLOWS_ENTRY):
            if (
                _get_ep_module(ep) == "orangecontrib.ewokstest.tutorials"
                and not is_ewokstest_category_enabled()
            ):
                continue
            yield ep

    tutorials_entry_points = examples_entry_points


def _get_ep_module(ep) -> str:
    try:
        return ep.module
    except AttributeError:
        return ep.module_name


def widgets_entry_points():
    return Config.widgets_entry_points()



import sysconfig

NAME = "Ewoks Test"

DESCRIPTION = "Short Test"

LONG_DESCRIPTION = "Long Test"

ICON = "icons/category.svg"

BACKGROUND = "light-blue"

WIDGET_HELP_PATH = (
    # Development documentation (make htmlhelp in ./doc)
    ("{DEVELOP_ROOT}/doc/_build/htmlhelp/index.html", None),
    # Documentation included in wheel
    ("{}/help/orange3-example/index.html".format(sysconfig.get_path("data")), None),
    # Online documentation url
    ("http://orange3-example-addon.readthedocs.io/en/latest/", ""),
)


def widget_discovery(discovery):
    """Do not show any widgets"""
    pass


def enable_ewokstest_category():
    global widget_discovery

    del widget_discovery


def is_ewokstest_category_enabled() -> bool:
    global widget_discovery
    try:
        widget_discovery(None)
    except NameError:
        return True
    return False



from ewoksorange.bindings import OWEwoksWidgetOneThreadPerRun
from ewoksorange.tests.examples.tasks import SumList2
from ewoksorange.tests.examples.widgets import WidgetMixin


class SumListSeveralThread(
    WidgetMixin, OWEwoksWidgetOneThreadPerRun, ewokstaskclass=SumList2
):
    """
    Simple demo class that create a new thread each time an execution of
    SumList2 is required
    """

    name = "SumList on several thread"
    description = "Sum all elements of a list using a new thread for each summation"
    icon = "icons/mywidget.svg"
    want_main_area = True



from enum import Enum

_OrangeVersion = Enum("OrangeVersion", "latest_orange oasys_fork latest_orange_base")

try:
    import oasys.widgets  # noqa F401

    ORANGE_VERSION = _OrangeVersion.oasys_fork
except ImportError:
    try:
        import Orange  # noqa F401
    except ImportError:
        ORANGE_VERSION = _OrangeVersion.latest_orange_base
    else:
        ORANGE_VERSION = _OrangeVersion.latest_orange



from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.canvas.widgetsscheme import (
        OASYSSignalManager as _SignalManagerWithSchemeOrg,
    )
    import oasys.canvas.widgetsscheme as widgetsscheme_module

    class _SignalManagerWithScheme(_SignalManagerWithSchemeOrg):
        def has_pending(self):
            return bool(self._input_queue)

    notify_input_helper = None
else:
    from orangewidget.workflow.widgetsscheme import (
        WidgetsSignalManager as _SignalManagerWithScheme,
    )
    import orangewidget.workflow.widgetsscheme as widgetsscheme_module

    from orangewidget.utils.signals import notify_input_helper

from ewokscore.variable import value_from_transfer

from .owwidgets import is_native_widget
from .qtapp import QtEvent
from ..bindings.owsignals import get_input_names, get_output_names
from . import invalid_data


class _MissingSignalValue:
    """Indicates a missing signal value and allows waiting for the real signal value"""

    completed = QtEvent()


class _SignalValues:
    """Store signal values per widget and allow waiting for a value"""

    def __init__(self):
        self.values = dict()

    def _get_value(self, widget, signal_name):
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        values = self.values.setdefault(widget, dict())
        if signal_name not in values:
            values[signal_name] = _MissingSignalValue()
        return values[signal_name]

    def _set_value(self, widget, signal_name, value):
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        values = self.values.setdefault(widget, dict())
        values[signal_name] = value

    def set_value(self, widget, signal_name, value):
        previous_value = self._get_value(widget, signal_name)
        self._set_value(widget, signal_name, value)
        if isinstance(previous_value, _MissingSignalValue):
            previous_value.completed.set()

    def invalidate_value(self, widget, signal_name):
        previous_value = self._get_value(widget, signal_name)
        if isinstance(previous_value, _MissingSignalValue):
            return
        self.set_value(widget, signal_name, _MissingSignalValue())

    def get_value(self, widget, signal_name, timeout=None):
        value = self._get_value(widget, signal_name)
        if isinstance(value, _MissingSignalValue):
            value.completed.wait(timeout=timeout)
            value = self._get_value(widget, signal_name)
        return value

    def has_value(self, widget, signal_name):
        value = self._get_value(widget, signal_name)
        return not isinstance(value, _MissingSignalValue)


class SignalManagerWithOutputTracking:
    """Store input and output signal value per widget. Knows
    when a widget is "executed" or not.

    Losely based on Orange.widgets.tests.base.DummySignalManager
    """

    def __init__(self, *args, **kwargs):
        self.output_values = _SignalValues()
        self.input_values = _SignalValues()
        self._widget_cache = dict()
        super().__init__(*args, **kwargs)

    def set_output_value(self, widget, signal_name, value):
        self.output_values.set_value(widget, signal_name, value)

    def invalidate_output_value(self, widget, signal_name):
        self.output_values.invalidate_value(widget, signal_name)

    def get_output_value(self, widget, signal_name, timeout=None):
        return self.output_values.get_value(widget, signal_name, timeout=timeout)

    def has_output_value(self, widget, signal_name):
        return self.output_values.has_value(widget, signal_name)

    def set_input_value(self, widget, signal_name, value):
        self.input_values.set_value(widget, signal_name, value)

    def invalidate_input_value(self, widget, signal_name):
        self.input_values.invalidate_value(widget, signal_name)

    def get_input_value(self, widget, signal_name, timeout=None):
        return self.input_values.get_value(widget, signal_name, timeout=timeout)

    def has_input_value(self, widget, signal_name):
        return self.input_values.has_value(widget, signal_name)

    def widget_is_executed(self, widget):
        variable_names, widget_has_outputs = self._widget_cache.get(
            widget, (None, None)
        )
        if variable_names is None:
            variable_names = get_output_names(widget)
            widget_has_outputs = variable_names
            if not widget_has_outputs:
                variable_names = get_input_names(widget)
            self._widget_cache[widget] = variable_names, widget_has_outputs
        if widget_has_outputs:
            # Widget is executed when all outputs are set
            for name in variable_names:
                if not self.has_output_value(widget, name):
                    return False
        else:
            # Widget has no outputs
            if variable_names:
                # Widget is executed when all inputs are set
                for name in variable_names:
                    if not self.has_input_value(widget, name):
                        return False
            else:
                # Widget is executed when any input is set
                return bool(self.input_values.values.get(widget, False))
        return True


class SignalManagerWithoutScheme(SignalManagerWithOutputTracking):
    """Used when no Orange canvas is present.

    Only needs to keep track of widget outputs because data between tasks
    is passed by the Ewoks mechanism, not the Orange mechanism (signal manager).
    """

    def send(self, widget, signal_name, value, *args, **kwargs):
        self.set_output_value(widget, signal_name, value)


class SignalManagerWithScheme(
    SignalManagerWithOutputTracking, _SignalManagerWithScheme
):
    """Used when the Orange canvas is present.

    Dereference `Variable` types for native Orange widget inputs.
    """

    def send(self, widget, signal_name, value, *args, **kwargs):
        super().send(widget, signal_name, value, *args, **kwargs)
        self.set_output_value(widget, signal_name, value)

    def process_signals_for_widget(self, node, widget, signals):
        for signal in signals:
            if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
                signal_name = signal.link.sink_channel
            else:
                signal_name = signal.channel.name
            self.set_input_value(widget, signal_name, signal.value)
        if is_native_widget(widget):
            modified_signals = list()
            for signal in signals:
                sinfo = signal._asdict()
                sinfo["value"] = value_from_transfer(sinfo["value"])
                modified_signals.append(type(signal)(**sinfo))
            signals = modified_signals
        super().process_signals_for_widget(node, widget, signals)

    def invalidate(self, node, channel):
        super().invalidate(node, channel)
        widget = self.scheme().widget_for_node(node)
        if widget is None:
            return
        self.invalidate_input_value(widget, channel.name, _MissingSignalValue())

    def widget_is_executed(self, widget):
        if self.has_pending():
            return False  # The widget might be executed again
        return super().widget_is_executed(widget)


def set_input_value(widget, signal, value, index):
    value = invalid_data.as_invalidation(value)
    key = id(widget), signal.name, signal.id
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        handler = getattr(widget, signal.handler)
        handler(value)
    else:
        notify_input_helper(signal, widget, value, key=key, index=index)


def patch_signal_manager():
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        widgetsscheme_module.OASYSSignalManager = SignalManagerWithScheme
    else:
        widgetsscheme_module.WidgetsSignalManager = SignalManagerWithScheme



from .bindings import load_graph  # noqa: F401
from .bindings import save_graph  # noqa: F401
from .bindings import convert_graph  # noqa: F401
from .bindings import execute_graph  # noqa: F401
from .bindings import graph_is_supported  # noqa: F401
from .bindings.owsignal_manager import patch_signal_manager
from .oasys_patch import oasys_patch

oasys_patch()
patch_signal_manager()

__version__ = "0.7.0rc"



import os
import logging
import numbers
from typing import Any, Callable, Dict, List, Sequence, Set, Union, Optional
from contextlib import contextmanager

from AnyQt import QtCore
from AnyQt import QtWidgets
from silx.gui.dialog.DataFileDialog import DataFileDialog
from ewokscore import missing_data

_logger = logging.getLogger(__name__)

ParameterValueType = Any
WidgetValueType = Union[str, numbers.Number, bool, missing_data.MissingData]


def default_serialize(value: ParameterValueType) -> WidgetValueType:
    return value


def default_deserialize(value: WidgetValueType) -> ParameterValueType:
    return value


@contextmanager
def block_signals(w: QtWidgets.QWidget):
    old = w.blockSignals(True)
    try:
        yield
    finally:
        w.blockSignals(old)


class ParameterForm(QtWidgets.QWidget):
    def __init__(self, *args, margin=0, spacing=4, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_ui(margin=margin, spacing=spacing)
        self._fields = dict()

    def _init_ui(self, margin=0, spacing=4):
        self._init_parent_ui()
        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(margin, margin, margin, margin)
        layout.setSpacing(spacing)
        self.setLayout(layout)

    def _init_parent_ui(self):
        parent = self.parent()
        if parent is None:
            return
        layout = parent.layout()
        if layout is None:
            layout = QtWidgets.QVBoxLayout()
            parent.setLayout(layout)
        layout.addWidget(self)

    def addParameter(
        self,
        name: str,
        value: ParameterValueType = missing_data.MISSING_DATA,
        value_for_type: Any = "",
        value_change_callback: Optional[Callable] = None,
        label: Optional[str] = None,
        readonly: Optional[bool] = None,
        enabled: Optional[bool] = None,
        select: Optional[str] = None,
        select_label: str = "...",
        checked: Optional[bool] = None,
        checkbox_label: str = "checked",
        serialize: Callable[[ParameterValueType], WidgetValueType] = default_serialize,
        deserialize: Callable[
            [WidgetValueType], ParameterValueType
        ] = default_deserialize,
        bool_label: str = "",
    ):
        """Each row has the following widgets:

        - label widget: name of the parameter by default
        - value widget: MISSING_DATA by default
        - select button (optional): select file/directory/HDF5 url
        - check box (optional): checked by default

        Parameters have folowing properties:

        - readonly: the means you can edit the value
                    Default: False when value_change_callback is provided, True otherwise
        - enabled: when False all widgets are disabled (grey color)
                   Default: True
        - checked: can mean whatever the user wants
        """
        if not label:
            label = name

        has_callback = bool(value_change_callback)
        if readonly is None:
            readonly = not has_callback
        if enabled is None:
            enabled = True

        _logger.debug(
            "Parameter form: initialize parameter %r (readonly = %s, enabled = %s, type = %s, missing = %s)",
            name,
            readonly,
            enabled,
            type(value_for_type),
            missing_data.is_missing_data(value),
        )

        label_widget = QtWidgets.QLabel(label)
        value_widget = None
        select_widget = None
        check_widget = None
        connections = list()

        if isinstance(value_for_type, str):
            value_widget = QtWidgets.QLineEdit()
            if value_change_callback:
                connections.append(
                    (value_widget.editingFinished.connect, value_change_callback)
                )
            if select == "file":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_file(
                            name,
                            must_exist=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "newfile":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_file(
                            name,
                            must_exist=False,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "directory":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_directory(
                            name, value_change_callback=value_change_callback
                        ),
                    )
                )
            elif select == "h5dataset":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_h5dataset(
                            name, value_change_callback=value_change_callback
                        ),
                    )
                )
            elif select == "h5group":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_h5group(
                            name, value_change_callback=value_change_callback
                        ),
                    )
                )
            elif select == "files":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_file(
                            name,
                            must_exist=True,
                            append=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "newfiles":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_file(
                            name,
                            must_exist=False,
                            append=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "directories":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_directory(
                            name,
                            append=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "h5datasets":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_h5dataset(
                            name,
                            append=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            elif select == "h5groups":
                select_widget = QtWidgets.QPushButton(select_label)
                connections.append(
                    (
                        select_widget.pressed.connect,
                        lambda: self._select_h5group(
                            name,
                            append=True,
                            value_change_callback=value_change_callback,
                        ),
                    )
                )
            else:
                select_widget = None
        elif isinstance(value_for_type, bool):
            value_widget = QtWidgets.QCheckBox(bool_label)
            if value_change_callback:
                connections.append(
                    (value_widget.stateChanged.connect, value_change_callback)
                )
        elif isinstance(value_for_type, numbers.Number):
            if isinstance(value_for_type, numbers.Integral):
                value_widget = QtWidgets.QSpinBox()
                value_widget.setRange(-(2**31), 2**31 - 1)
            else:
                value_widget = QtWidgets.QDoubleSpinBox()
                value_widget.setRange(-(2**52), 2**52 - 1)
            if value_change_callback:
                connections.append(
                    (value_widget.editingFinished.connect, value_change_callback)
                )
        elif isinstance(value_for_type, Sequence):
            value_widget = QtWidgets.QComboBox()
            value_widget.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
            value_widget.addItem("<missing>", missing_data.MISSING_DATA)
            for data in value_for_type:
                data = serialize(data)
                value_widget.addItem(str(data), data)
            if value_change_callback:
                connections.append(
                    (value_widget.currentIndexChanged.connect, value_change_callback)
                )
        else:
            raise TypeError(
                f"Parameter '{name}' with type '{type(value)}' does not have a Qt widget"
            )

        if checked is not None:
            check_widget = QtWidgets.QCheckBox(checkbox_label)
            check_widget.setChecked(checked)
            if value_change_callback:
                connections.append(
                    (check_widget.stateChanged.connect, value_change_callback)
                )

        policy = QtWidgets.QSizePolicy.Expanding
        value_widget.setSizePolicy(policy, policy)
        grid = self.layout()
        row = grid.rowCount()
        grid.addWidget(label_widget, row, 0)
        if value_widget:
            grid.addWidget(value_widget, row, 1)
        if select_widget:
            grid.addWidget(select_widget, row, 2)
        if check_widget:
            grid.addWidget(check_widget, row, 3)

        self._fields[name] = {
            "row": row,
            "deserialize": deserialize,
            "serialize": serialize,
        }

        self.set_parameter_value(name, value)
        self.set_parameter_readonly(name, readonly)
        self.set_parameter_enabled(name, enabled)
        for connect, func in connections:
            connect(func)

    def _get_widget(self, name: str, col: int) -> QtWidgets.QWidget:
        if name not in self._fields:
            return None
        row = self._fields[name]["row"]
        item = self.layout().itemAtPosition(row, col)
        if item is None:
            return None
        return item.widget()

    def _get_label_widget(self, name: str) -> QtWidgets.QWidget:
        return self._get_widget(name, 0)

    def _get_value_widget(self, name: str) -> QtWidgets.QWidget:
        return self._get_widget(name, 1)

    def _get_select_widget(self, name: str) -> QtWidgets.QWidget:
        return self._get_widget(name, 2)

    def _get_check_widget(self, name: str) -> QtWidgets.QWidget:
        return self._get_widget(name, 3)

    def has_parameter(self, name: str):
        w = self._get_value_widget(name)
        return w is not None

    def get_parameter_value(self, name: str):
        w = self._get_value_widget(name)
        if w is None:
            return missing_data.MISSING_DATA

        if isinstance(w, QtWidgets.QLineEdit):
            wvalue = w.text()
        elif isinstance(w, QtWidgets.QCheckBox):
            wvalue = w.isChecked()
        elif isinstance(w, QtWidgets.QComboBox):
            wvalue = w.currentData()
            # TODO: check what it returns when index is -1
        else:
            wvalue = w.value()

        fdict = self._fields[name]
        changed = wvalue != fdict["last_widget_value"]
        if changed:
            fdict["last_widget_value"] = wvalue
        elif fdict["null"]:
            return missing_data.MISSING_DATA

        deserialize = fdict["deserialize"]
        try:
            value = deserialize(wvalue)
        except Exception:
            value = missing_data.MISSING_DATA
            fdict["null"] = True
        else:
            fdict["null"] = missing_data.is_missing_data(value)

        return value

    def set_parameter_value(self, name: str, value: ParameterValueType):
        """When the value cannot be serialized or cannot be set to the widget,
        the parameter value will set to MISSING_DATA.
        """
        w = self._get_value_widget(name)
        if w is None:
            return

        fdict = self._fields[name]
        serialize = fdict["serialize"]
        if missing_data.is_missing_data(value):
            set_value = missing_data.MISSING_DATA
        else:
            try:
                set_value = serialize(value)
            except Exception:
                set_value = missing_data.MISSING_DATA

        if isinstance(w, QtWidgets.QLineEdit):
            _logger.debug("Parameter form: set string parameter %r = %r", name, value)
            try:
                with block_signals(w):
                    w.setText(set_value)
            except TypeError:
                with block_signals(w):
                    w.setText("")
                fdict["null"] = True
                fdict["last_widget_value"] = ""
            else:
                fdict["null"] = False
                fdict["last_widget_value"] = set_value
        elif isinstance(w, QtWidgets.QCheckBox):
            _logger.debug("Parameter form: set boolean parameter %r = %r", name, value)
            try:
                with block_signals(w):
                    w.setChecked(set_value)
            except TypeError:
                with block_signals(w):
                    w.setChecked(False)
                fdict["null"] = True
                fdict["last_widget_value"] = False
            else:
                fdict["null"] = False
                fdict["last_widget_value"] = set_value
        elif isinstance(w, QtWidgets.QComboBox):
            _logger.debug(
                "Parameter form: select choice parameter %r = %r", name, value
            )
            idx = w.findData(set_value)
            if idx == -1:
                with block_signals(w):
                    w.setCurrentIndex(0)
                fdict["null"] = True
                fdict["last_widget_value"] = missing_data.MISSING_DATA
            else:
                with block_signals(w):
                    w.setCurrentIndex(idx)
                fdict["null"] = False
                fdict["last_widget_value"] = set_value
        else:
            _logger.debug(
                "Parameter form: set numerical parameter %r = %r", name, value
            )
            try:
                with block_signals(w):
                    w.setValue(set_value)
            except TypeError:
                with block_signals(w):
                    w.setValue(0)
                fdict["null"] = True
                fdict["last_widget_value"] = 0
            else:
                fdict["null"] = False
                fdict["last_widget_value"] = set_value

    def get_parameter_readonly(self, name: str) -> Optional[bool]:
        w = self._get_value_widget(name)
        if w is not None:
            if isinstance(w, QtWidgets.QCheckBox):
                return w.isEnabled()
            elif isinstance(w, QtWidgets.QComboBox):
                return w.isEnabled()
            else:
                return w.isReadOnly()

    def set_parameter_readonly(self, name: str, value: bool) -> None:
        w = self._get_value_widget(name)
        if w is not None:
            if isinstance(w, QtWidgets.QCheckBox):
                return w.setEnabled(value)
            elif isinstance(w, QtWidgets.QComboBox):
                return w.setEnabled(value)
            else:
                return w.setReadOnly(value)

    def get_parameter_enabled(self, name: str) -> Optional[bool]:
        w = self._get_value_widget(name)
        if w is not None:
            return w.isEnabled()

    def set_parameter_enabled(self, name: str, value: bool) -> None:
        w = self._get_value_widget(name)
        if w is not None:
            w.setEnabled(value)

    def get_parameter_checked(self, name: str) -> Optional[bool]:
        w = self._get_check_widget(name)
        if w is not None:
            return w.isChecked()

    def set_parameter_checked(self, name: str, value: bool) -> None:
        w = self._get_check_widget(name)
        if w is not None:
            w.setChecked(value)

    def get_parameter_names(self) -> Set[str]:
        return set(self._fields)

    def get_parameter_values(self) -> Dict[str, ParameterValueType]:
        return {name: self.get_parameter_value(name) for name in self._fields}

    def set_parameter_values(self, params: Dict[str, ParameterValueType]) -> None:
        for name, value in params.items():
            self.set_parameter_value(name, value)

    def get_parameters_readonly(self) -> Dict[str, bool]:
        return {name: self.get_parameter_readonly(name) for name in self._fields}

    def set_parameters_readonly(self, params: Dict[str, bool]) -> None:
        for name, value in params.items():
            self.set_parameter_readonly(name, value)

    def get_parameters_checked(self) -> Dict[str, bool]:
        result = dict()
        for name in self._fields:
            checked = self.get_parameter_checked(name)
            if checked is None:
                continue
            result[name] = checked
        return result

    def set_parameters_checked(self, params: Dict[str, bool]) -> None:
        for name, value in params.items():
            self.set_parameter_checked(name, value)

    def get_parameters_enabled(self) -> Dict[str, bool]:
        return {name: self.get_parameter_enabled(name) for name in self._fields}

    def set_parameters_enabled(self, params: Dict[str, bool]) -> None:
        for name, value in params.items():
            self.set_parameter_enabled(name, value)

    def _select_file(
        self,
        name: str,
        must_exist: bool = True,
        append: bool = False,
        value_change_callback=Optional[Callable[[], None]],
    ) -> None:
        if append:
            filename = self._list_value_first(name)
        else:
            filename = self.get_parameter_value(name)
        dialog = QtWidgets.QFileDialog(self)
        if must_exist:
            dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        else:
            dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
        if filename:
            dialog.setDirectory(os.path.dirname(filename))

        if not dialog.exec_():
            dialog.close()
            return

        value = dialog.selectedFiles()[0]
        if append:
            value = self._list_value_append(name, value)
        self.set_parameter_value(name, value)
        if value_change_callback:
            value_change_callback()

    def _select_h5dataset(
        self,
        name: str,
        append: bool = False,
        value_change_callback=Optional[Callable[[], None]],
    ) -> None:
        if append:
            url = self._list_value_first(name)
        else:
            url = self.get_parameter_value(name)
        dialog = DataFileDialog(self)
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingDataset)
        if url:
            dialog.selectUrl(url)

        if not dialog.exec():
            dialog.close()
            return

        value = dialog.selectedUrl()
        if value:
            value = value.replace("?/", "?path=/")
        if append:
            value = self._list_value_append(name, value)
        self.set_parameter_value(name, value)
        if value_change_callback:
            value_change_callback()

    def _select_h5group(
        self,
        name: str,
        append: bool = False,
        value_change_callback=Optional[Callable[[], None]],
    ) -> None:
        if append:
            url = self._list_value_first(name)
        else:
            url = self.get_parameter_value(name)
        dialog = DataFileDialog(self)
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingGroup)
        if url:
            dialog.selectUrl(url)

        if not dialog.exec():
            dialog.close()
            return

        value = dialog.selectedUrl()
        if value:
            value = value.replace("?/", "?path=/")
        if append:
            value = self._list_value_append(name, value)
        self.set_parameter_value(name, value)
        if value_change_callback:
            value_change_callback()

    def _select_directory(
        self,
        name: str,
        append: bool = False,
        value_change_callback=Optional[Callable[[], None]],
    ) -> None:
        if append:
            directory = self._list_value_first(name)
        else:
            directory = self.get_parameter_value(name)
        dialog = QtWidgets.QFileDialog(self)
        dialog.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
        if directory:
            dialog.setDirectory(directory)

        if not dialog.exec_():
            dialog.close()
            return

        value = dialog.selectedFiles()[0]
        if append:
            value = self._list_value_append(name, value)
        self.set_parameter_value(name, value)
        if value_change_callback:
            value_change_callback()

    def _list_value_append(self, name: str, value: str) -> List[str]:
        lst = self.get_parameter_value(name)
        if not lst:
            lst = [value]
        elif isinstance(lst, str):
            lst = [lst, value]
        elif isinstance(lst, list):
            lst.append(value)
        else:
            raise TypeError(value)
        return lst

    def _list_value_first(self, name: str) -> Optional[str]:
        lst = self.get_parameter_value(name)
        if not lst:
            return None
        elif isinstance(lst, str):
            return lst
        elif isinstance(lst, list):
            return lst[-1]
        else:
            raise TypeError(lst)

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Enter or key == QtCore.Qt.Key_Return:
            # TODO: Orange3 causes a button in focus to be pressed due to this event.
            return
        super().keyPressEvent(event)



import gc
import logging
import pytest
from ewoksorange.bindings.qtapp import qtapp_context
from ewoksorange.bindings.qtapp import get_all_qtwidgets
from ewoksorange.canvas.handler import OrangeCanvasHandler
from ewoksorange.orange_version import ORANGE_VERSION
from orangecontrib.ewokstest import enable_ewokstest_category

logger = logging.getLogger(__name__)


def global_cleanup_orange():
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        pass
    else:
        from orangecanvas.document.suggestions import Suggestions

        Suggestions.instance = None


def global_cleanup_pytest():
    for obj in gc.get_objects():
        if isinstance(obj, logging.LogRecord):
            obj.exc_info = None  # traceback keeps frames which keep locals


def collect_garbage(app):
    app.processEvents()
    while gc.collect():
        app.processEvents()


@pytest.fixture(scope="session")
def qtapp():
    enable_ewokstest_category()
    with qtapp_context() as app:
        assert app is not None
        yield app
    collect_garbage(app)
    global_cleanup_orange()
    global_cleanup_pytest()
    collect_garbage(app)
    warn_qtwidgets_alive()


@pytest.fixture(scope="session")
def raw_ewoks_orange_canvas(qtapp):
    with OrangeCanvasHandler() as handler:
        yield handler


@pytest.fixture()
def ewoks_orange_canvas(raw_ewoks_orange_canvas):
    yield raw_ewoks_orange_canvas
    try:
        raw_ewoks_orange_canvas.scheme.ewoks_finalize()
    except AttributeError:
        pass


def warn_qtwidgets_alive():
    widgets = get_all_qtwidgets()
    if widgets:
        logger.warning("%d remaining widgets after tests", len(widgets))



import logging
from typing import Optional, Type
from AnyQt.QtCore import QThread
from ewokscore.task import Task
from ewokscore.task import TaskInputError
from ewokscore import TaskWithProgress


_logger = logging.getLogger(__name__)


class TaskExecutor:
    """Create and execute an Ewoks task"""

    def __init__(self, ewokstaskclass: Type[Task]) -> None:
        self.__ewokstaskclass = ewokstaskclass
        self.__task = None
        self.__task_init_exception = None

    def create_task(self, log_missing_inputs: bool = False, **kwargs) -> None:
        if not issubclass(self.__ewokstaskclass, TaskWithProgress):
            kwargs.pop("progress", None)
        self.__task = None
        self.__task_init_exception = None
        try:
            self.__task = self.__ewokstaskclass(**kwargs)
        except TaskInputError as e:
            self.__task_init_exception = e
            if log_missing_inputs:
                _logger.error(f"task initialization failed: {e}", exc_info=True)
            else:
                _logger.info(f"task initialization failed: {e}")

    def execute_task(self) -> None:
        if not self.has_task:
            return
        try:
            self.__task.execute()
        except Exception as e:
            _logger.error(f"task failed: {e}", exc_info=True)

    @property
    def has_task(self) -> bool:
        return self.__task is not None

    @property
    def succeeded(self) -> Optional[bool]:
        """Returns `None` when the task was not or could not be instantiated"""
        if self.__task is None:
            return None
        return self.__task.succeeded

    @property
    def done(self) -> Optional[bool]:
        """Returns `None` when the task was not or could not be instantiated"""
        if self.__task is None:
            return None
        return self.__task.done

    @property
    def exception(self) -> Optional[Exception]:
        """The instantiation exception, the execution exception or `None`"""
        if self.__task_init_exception is not None:
            return self.__task_init_exception
        if self.__task is None:
            return None
        return self.__task.exception

    @property
    def output_variables(self) -> Optional[dict]:
        if self.__task is None:
            return dict()
        return self.__task.output_variables

    @property
    def current_task(self) -> Optional[Task]:
        return self.__task


class ThreadedTaskExecutor(QThread, TaskExecutor):
    """Create and execute an Ewoks task in a dedicated thread."""

    def run(self) -> None:
        self.execute_task()

    def stop(self, timeout: Optional[float] = None, wait: bool = False) -> None:
        self.blockSignals(True)
        if wait:
            if timeout:
                self.wait(timeout * 1000)
            else:
                self.wait()
        if self.isRunning():
            self.quit()



from ewoksorange.gui.parameterform import ParameterForm
from ewoksorange.gui.orange_imports import gui


class SimpleTypesWidgetMixin:
    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        box = gui.widgetBox(self.controlArea, "Default Inputs")
        self._default_inputs_form = ParameterForm(parent=box)

        for name, value in self.get_default_input_values(include_missing=True).items():
            options = self._get_parameter_options(name)
            self._default_inputs_form.addParameter(
                name,
                value=value,
                value_change_callback=self._default_inputs_changed,
                **options
            )

        box = gui.widgetBox(self.controlArea, "Inputs From Previous Task")
        self._dynamic_input_form = ParameterForm(parent=box)
        for name in self.get_input_names():
            options = self._get_parameter_options(name)
            self._dynamic_input_form.addParameter(
                name, readonly=True, enabled=False, **options
            )

    def _init_main_area(self):
        super()._init_main_area()
        box = gui.widgetBox(self.mainArea, "Outputs")
        self._output_form = ParameterForm(parent=box)
        for name in self.get_output_names():
            options = self._get_parameter_options(name)
            self._output_form.addParameter(name, readonly=True, **options)

    def _get_parameter_options(self, name):
        return {}

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())

    def handleNewSignals(self):
        names = set(self.get_input_names())
        for name, value in self.get_dynamic_input_values().items():
            names.remove(name)
            self._dynamic_input_form.set_parameter_enabled(name, True)
            self._default_inputs_form.set_parameter_enabled(name, False)
            self._dynamic_input_form.set_parameter_value(name, value)
        for name in names:
            self._dynamic_input_form.set_parameter_enabled(name, False)
            self._default_inputs_form.set_parameter_enabled(name, True)
        super().handleNewSignals()

    def task_output_changed(self):
        for name, value in self.get_task_output_values().items():
            self._output_form.set_parameter_value(name, value)
        super().task_output_changed()


class IntegerAdderMixin(SimpleTypesWidgetMixin):
    def _get_parameter_options(self, name):
        return {"value_for_type": 0}



import inspect
from typing import Callable, List, Optional
from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from collections import namedtuple

    Input = namedtuple("Input", ["name", "type", "handler"])
    Input.ewoksname = property(lambda self: self.name)

    class Output(dict):
        @property
        def ewoksname(self):
            return self["name"]

        def __getattr__(self, attr):
            return self[attr]

else:
    from orangewidget.widget import Input, Output


SIGNAL_TYPES = (Input, Output)


def is_signal(obj):
    return isinstance(obj, SIGNAL_TYPES)


def get_signals(signal_container, orange_to_ewoks: Optional[dict] = None) -> dict:
    """Returns a map from ewoks names to signal objects"""
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        if not signal_container:
            return dict()
        if isinstance(signal_container[0], tuple):
            return {
                orange_to_ewoks.get(tpl[0], tpl[0]): tpl for tpl in signal_container
            }
        else:
            return {
                orange_to_ewoks.get(adict["name"], adict["name"]): adict
                for adict in signal_container
            }
    else:
        """Returns a map from ewoks names to signal objects"""
        # TODO: getsignals doesn't work in the Orange3 hard-fork
        # from orangewidget.utils.signals import getsignals
        # return dict(getsignals(signal_container))
        return dict(inspect.getmembers(signal_container, is_signal))


def get_signal_container(widget_class, direction: str):
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        return getattr(widget_class, direction, list())
        # List of tuples or dictionaries
    else:
        return getattr(widget_class, direction.title(), type)
        # Class


def get_orange_to_ewoks_mapping(widget_class, direction: str) -> Optional[dict]:
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        return getattr(widget_class, direction + "_orange_to_ewoks", dict())
    else:
        return None


def get_ewoks_to_orange_mapping(widget_class, direction: str) -> Optional[dict]:
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        return {
            v: k
            for k, v in get_orange_to_ewoks_mapping(widget_class, direction).items()
        }
    else:
        return None


def signal_ewoks_to_orange_name(widget_class, direction: str, ewoksname: str) -> str:
    signal_container = get_signal_container(widget_class, direction)
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        orange_to_ewoks = get_orange_to_ewoks_mapping(widget_class, direction)
        for signal in signal_container:
            signal_ewoksname = orange_to_ewoks.get(signal.name, signal.name)
            if signal_ewoksname == ewoksname:
                return signal.name
        raise RuntimeError(f"{ewoksname} is not a signal of {signal_container}")
    else:
        try:
            return getattr(signal_container, ewoksname).name
        except AttributeError:
            raise RuntimeError(
                f"{ewoksname} is not a signal of {signal_container} of {widget_class}"
            ) from None


def signal_orange_to_ewoks_name(widget_class, direction: str, orangename: str) -> str:
    signal_container = get_signal_container(widget_class, direction)
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        orange_to_ewoks = get_orange_to_ewoks_mapping(widget_class, direction)
        for signal in signal_container:
            if signal.name == orangename:
                return orange_to_ewoks.get(orangename, orangename)
        raise RuntimeError(f"{orangename} is not a signal of {signal_container}")
    else:
        for ewoksname, signal in get_signals(signal_container).items():
            if signal.name == orangename:
                return ewoksname
        raise RuntimeError(f"{orangename} is not a signal of {signal_container}")


def receive_dynamic_input(name: str) -> Callable:
    setter_name = f"{name}_ewoks_setter"

    def setter(self, value):
        # Called by the SignalManager as a result of calling
        # `send` on an upstream output.
        self.set_dynamic_input(name, value)

    setter.__name__ = setter_name
    return setter


def _validate_signals_oasys(namespace: dict, direction: str, names: List[str]) -> None:
    is_inputs = direction == "inputs"
    signal_container = namespace[direction]

    orange_to_ewoks = namespace.get(direction + "_orange_to_ewoks", dict())
    signal_dict = get_signals(signal_container, orange_to_ewoks)

    signal_container = namespace[direction] = list()
    for ewoksname in names:
        signal = signal_dict.get(ewoksname, None)
        if is_inputs:
            if signal is None:
                orangename, stype, handler = ewoksname, object, None
            else:
                orangename, stype, handler = signal

            if not handler:
                setter = receive_dynamic_input(ewoksname)
                namespace[setter.__name__] = setter
                handler = setter.__name__
            signal = Input(
                name=orangename,
                type=stype,
                handler=handler,
            )
        else:
            if signal is None:
                signal = Output(
                    [("name", ewoksname), ("type", object), ("id", ewoksname)]
                )
            else:
                signal = Output(signal)
        signal_container.append(signal)


def _validate_signals(namespace: dict, direction: str, names: List[str]) -> None:
    """
    convert ewoks inputs or outputs to orange Input and or Output
    """
    signals_class_name = direction.title()
    is_inputs = direction == "inputs"
    signal_container = namespace[signals_class_name]

    signal_dict = get_signals(signal_container)

    for ewoksname in names:
        signal = signal_dict.get(ewoksname, None)
        if signal is None:
            if is_inputs:
                signal = Input(name=ewoksname, type=object)
            else:
                signal = Output(name=ewoksname, type=object)
            setattr(signal_container, ewoksname, signal)
        signal.ewoksname = ewoksname
        if is_inputs and not signal.handler:  # str
            setter = receive_dynamic_input(ewoksname)
            namespace[setter.__name__] = signal(setter)


def validate_inputs(namespace, name_to_ignore=tuple()) -> None:
    """Adds missing Orange inputs by compaing the existing Orange
    inputs with the ewoks inputs."""
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        if "inputs" not in namespace:
            namespace["inputs"] = list()

        if "inputs_orange_to_ewoks" not in namespace:
            namespace["inputs_orange_to_ewoks"] = dict()

        _validate_signals_oasys(
            namespace, "inputs", namespace["ewokstaskclass"].input_names()
        )
    else:
        if "Inputs" not in namespace:

            class Inputs:
                pass

            namespace["Inputs"] = Inputs

        _validate_signals(
            namespace=namespace,
            direction="inputs",
            names=tuple(
                filter(
                    lambda name: name not in name_to_ignore,
                    namespace["ewokstaskclass"].input_names(),
                )
            ),
        )


def validate_outputs(namespace, name_to_ignore=tuple()) -> None:
    """Adds missing Orange outputs by compaing the existing Orange
    outputs with the ewoks outputs."""
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        if "outputs" not in namespace:
            namespace["outputs"] = list()

        if "outputs_orange_to_ewoks" not in namespace:
            namespace["outputs_orange_to_ewoks"] = dict()

        _validate_signals_oasys(
            namespace, "outputs", namespace["ewokstaskclass"].output_names()
        )
    else:
        if "Outputs" not in namespace:

            class Outputs:
                pass

            namespace["Outputs"] = Outputs

        _validate_signals(
            namespace=namespace,
            direction="outputs",
            names=tuple(
                filter(
                    lambda name: name not in name_to_ignore,
                    namespace["ewokstaskclass"].output_names(),
                )
            ),
        )


def get_input_names(widget_class) -> List[str]:
    """Returns the Orange signal names, not the Ewoks output names"""
    signal_container = get_signal_container(widget_class, "inputs")
    orange_to_ewoks = get_orange_to_ewoks_mapping(widget_class, "inputs")
    return [
        signal.name
        for signal in get_signals(signal_container, orange_to_ewoks).values()
    ]


def get_output_names(widget_class) -> List[str]:
    """Returns the Orange signal names, not the Ewoks output names"""
    signal_container = get_signal_container(widget_class, "outputs")
    orange_to_ewoks = get_orange_to_ewoks_mapping(widget_class, "outputs")
    return [
        signal.name
        for signal in get_signals(signal_container, orange_to_ewoks).values()
    ]



"""Commonly used Orange3 components for implementing Orange Widgets.
"""

from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets import gui  # noqa F401
    from orangewidget.settings import Setting  # noqa F401

    Input = None
    Output = None
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets import gui  # noqa F401
    from orangewidget.widget import Input, Output  # noqa F401
    from orangewidget.settings import Setting  # noqa F401
else:
    from orangewidget import gui  # noqa F401
    from orangewidget.widget import Input, Output  # noqa F401
    from orangewidget.settings import Setting  # noqa F401
